from typing import List, Union

from statham.dsl.constants import Maybe
from statham.dsl.elements import (
    AnyOf,
    Array,
    Boolean,
    Integer,
    Null,
    Number,
    OneOf,
    Object,
    ObjectOptions,
    String,
)
from statham.dsl.property import Property


class Model(Object):

    list_of_strings: Maybe[List[str]] = Property(Array(String()))



from typing import Any, Optional

from statham.dsl.constants import NotPassed
from statham.dsl.exceptions import ValidationError
from statham.dsl.validation.base import Validator


class Required(Validator):
    types = (dict,)
    keywords = ("required",)
    message = "Must contain all required fields: {required}"

    def validate(self, value: Any):
        if set(self.params["required"]) - set(value):
            raise ValidationError


class AdditionalProperties(Validator):
    types = (dict,)
    keywords = ("properties", "additionalProperties")
    message = "Must not contain unspecified properties. Accepts: {properties}"

    @classmethod
    def from_element(cls, element) -> Optional["AdditionalProperties"]:
        params = list(
            getattr(element, keyword, None) for keyword in cls.keywords
        )
        if None in params:
            return None
        params[0] = set(params[0] or {})
        params[1] = bool(
            True if isinstance(params[1], NotPassed) else params[1]
        )
        return cls(*params)

    def validate(self, value: Any):
        if self.params["additionalProperties"]:
            return
        if set(value) - self.params["properties"]:
            raise ValidationError



from statham.dsl.elements import Element
from statham.orderer import Orderer


# TODO: Identify which of these to exclude when performing generation.
_IMPORT_STATEMENTS = """from typing import List, Union

from statham.dsl.constants import Maybe
from statham.dsl.elements import (
    AnyOf,
    Array,
    Boolean,
    Integer,
    Null,
    Number,
    OneOf,
    Object,
    ObjectOptions,
    String,
)
from statham.dsl.property import Property


"""


def serialize_python(*elements: Element) -> str:
    """Output python declaration code.

    Captures declaration of the first Object elements, and any subsequent
    elements this depends on.
    """
    return _IMPORT_STATEMENTS + "\n\n".join(
        [object_model.python() for object_model in Orderer(*elements)]
    )



from collections import defaultdict
import inspect
from itertools import chain
import re
import string
from typing import Any, Callable, DefaultDict, Dict, List, Type, Union
import unicodedata

from statham.dsl.constants import COMPOSITION_KEYWORDS, NotPassed
from statham.dsl.elements import (
    AllOf,
    AnyOf,
    Array,
    Boolean,
    CompositionElement,
    Integer,
    Null,
    Number,
    Object,
    ObjectOptions,
    OneOf,
    Element,
    String,
)
from statham.dsl.elements.meta import (
    ObjectClassDict,
    ObjectMeta,
    RESERVED_PROPERTIES,
)
from statham.dsl.exceptions import FeatureNotImplementedError, SchemaParseError
from statham.dsl.helpers import expand, reraise, split_dict
from statham.dsl.property import _Property


_TYPE_MAPPING = {
    "array": Array,
    "boolean": Boolean,
    "integer": Integer,
    "null": Null,
    "number": Number,
    "string": String,
}


class ParseState:
    """Recusive state.

    Used to de-duplicate models which are traversed multiple times, and to
    rename distinct models with the same name.
    """

    def __init__(self):
        self.seen: DefaultDict[str, List[ObjectMeta]] = defaultdict(list)

    def dedupe(self, object_type: ObjectMeta):
        """Deduplicate a parsed model.

        If it has been seen before, then return the existing one. Otherwise
        ensure the model's name is distinct from other models and keep store
        it.
        """
        name = object_type.__name__
        for existing in self.seen[name]:
            if object_type == existing:
                return existing
        count = len(self.seen[name])
        if count:
            object_type.__name__ = name + f"_{count}"
        self.seen[name].append(object_type)
        return object_type


def parse(schema: Dict[str, Any]) -> List[Element]:
    """Parse a JSONSchema document to DSL Element format.

    Checks the top-level and definitions keyword to collect elements.
    """
    state = ParseState()
    return [parse_element(schema, state)] + [
        parse_element(definition, state)
        for definition in schema.get("definitions", {}).values()
        if isinstance(definition, (dict, Element))
    ]


@reraise(
    RecursionError,
    SchemaParseError,
    "Could not parse cyclical dependencies of this schema.",
)
def parse_element(schema: Dict[str, Any], state: ParseState = None) -> Element:
    """Parse a JSONSchema element to a DSL Element object."""
    state = state or ParseState()
    if isinstance(schema, Element):
        return schema
    if "properties" in schema:
        schema["properties"] = parse_properties(schema, state)
    if "items" in schema:
        schema["items"] = parse_items(schema, state)
    schema["additionalProperties"] = parse_additional_properties(schema, state)
    if {"anyOf", "oneOf", "allOf"} & set(schema):
        return parse_composition(schema, state)
    if "type" not in schema:
        return Element(**keyword_filter(Element)(schema))
    return parse_typed(schema["type"], schema, state)


def parse_composition(
    schema: Dict[str, Any], state: ParseState = None
) -> Element:
    """Parse a schema with composition keywords.

    Handles multiple composition keywords by wrapping them in an AllOf
    element. Similarly, non-keyword elements are parsed as usual and
    included in an allOf element.

    For example:
    ```python
    schema = {
        "type": "integer",
        "oneOf": [{"minimum": 3}, {"maximum": 5}],
        "anyOf": [{"multipleOf": 2}, {"multipleOf": 3}],
    }
    parse_composition(schema) == AllOf(
        Integer(),
        OneOf(Element(minimum=3), Element(maximum=5)),
        AnyOf(Element(multipleOf=2), Element(multipleOf=3)),
    )
    ```
    """
    state = state or ParseState()
    composition, other = split_dict(set(COMPOSITION_KEYWORDS) | {"default"})(
        schema
    )
    base_element = parse_element(other, state)
    for key in COMPOSITION_KEYWORDS:
        composition[key] = [
            parse_element(sub_schema) for sub_schema in composition.get(key, [])
        ]
    all_of = [base_element] + composition["allOf"]
    all_of.append(_compose_elements(OneOf, composition["oneOf"]))
    all_of.append(_compose_elements(AnyOf, composition["anyOf"]))
    element = _compose_elements(AllOf, all_of)
    element.default = composition.get("default") or element.default
    return element


def parse_typed(
    type_value: Any, schema: Dict[str, Any], state: ParseState = None
) -> Element:
    """Parse a typed schema with no composition keywords.

    :raises KeyError: if no "type" key is present.
    :raises SchemaParseError: if value at "type" is not a `str` or `list`.
    """
    state = state or ParseState()
    if not isinstance(type_value, (str, list)):
        raise SchemaParseError.invalid_type(type_value)
    if isinstance(type_value, list):
        return parse_multi_typed(type_value, schema, state)
    if schema["type"] == "object":
        return parse_object(schema, state)
    if schema["type"] == "array":
        return parse_array(schema, state)
    element_type = _TYPE_MAPPING[type_value]
    sub_schema = keyword_filter(element_type)(schema)
    return element_type(**sub_schema)


def parse_multi_typed(
    type_list: List[str], schema: Dict[str, Any], state: ParseState = None
) -> CompositionElement:
    """Parse a schema with multiple type values.

    Converts schema to an equivalent representation using "anyOf". For example:
    ```python
    {"type": ["string", "integer"]}
    ```
    becomes
    ```
    {"anyOf": [{"type": "string"}, {"type": "integer"}]}
    ```
    """
    state = state or ParseState()
    default = schema.get("default", NotPassed())
    schema = {key: val for key, val in schema.items() if key != "default"}
    if len(type_list) == 1:
        return parse_element({**schema, "type": type_list[0]}, state)
    return AnyOf(
        *(
            parse_element({**schema, "type": type_value}, state)
            for type_value in type_list
        ),
        default=default,
    )


def parse_object(
    schema: Dict[str, Any], state: ParseState = None
) -> ObjectMeta:
    """Parse an object schema element to an `Object` subclass.

    The name of the generated class is derived from the following keys
    in precedence order:
    - "title"
    - "_x_autotitle"

    Also requires that "properties" and "additionalProperties" values have
    already been parsed.

    :raises SchemaParseError: if keys exist from which to derive the class
        title.
    """
    state = state or ParseState()
    title = schema.get("title", schema.get("_x_autotitle"))
    if not title:
        raise SchemaParseError.missing_title(schema)
    title = _title_format(title)
    default = schema.get("default", NotPassed())
    properties = schema.get("properties", {})
    properties.update(
        {
            key: _Property(Element(), required=True)
            for key in schema.get("required", [])
            if key not in properties
        }
    )
    class_dict = ObjectClassDict(
        default=default,
        options=ObjectOptions(**keyword_filter(ObjectOptions)(schema)),
    )
    for key, value in properties.items():
        class_dict[key] = value
    object_type = ObjectMeta(title, (Object,), class_dict)
    return state.dedupe(object_type)


def parse_properties(
    schema: Dict[str, Any], state: ParseState = None
) -> Dict[str, _Property]:
    """Parse properties from a schema element."""
    state = state or ParseState()
    required = set(schema.get("required", []))
    properties = schema.get("properties", {})
    return {
        **{
            parse_attribute_name(key): _Property(
                parse_element(value, state),
                required=key in required,
                source=key,
            )
            for key, value in properties.items()
            # Ignore malformed values.
            if isinstance(value, dict)
        },
        **{
            parse_attribute_name(key): prop
            for key, prop in properties.items()
            if isinstance(prop, _Property)
        },
    }


def parse_attribute_name(name: str) -> str:
    """Convert attibute name to valid python attribute.

    Attempts to replace special characters with their unicode names.
    """

    def _char_map(idx: int, char: str) -> str:
        if char.isalnum() or char in ("_", "-", " "):
            return char
        if char in string.whitespace:
            return "_"
        label = unicodedata.name(char, "unknown").lower()
        if idx != 0 and name[idx - 1] != "_":
            label = "_" + label
        if idx != len(name) - 1 and name[idx + 1] != "_":
            label = label + "_"
        return label

    chars = map(expand(_char_map), enumerate(name))
    name = "".join(chars).replace(" ", "_").replace("-", "_")
    if not name:
        return "blank"
    return (
        name
        if name not in RESERVED_PROPERTIES and name[0].isalpha()
        else f"_{name}"
    )


def parse_additional_properties(
    schema: Dict[str, Any], state: ParseState = None
) -> Union[Element, bool]:
    """Parse additionalProperties from a schema element.

    If key is not present, defaults to `True`.
    """
    state = state or ParseState()
    additional_properties = schema.get("additionalProperties", True)
    if isinstance(additional_properties, bool):
        return additional_properties
    return parse_element(additional_properties, state)


def parse_array(schema: Dict[str, Any], state: ParseState = None) -> Array:
    """Parse an array schema element."""
    state = state or ParseState()
    items = schema.get("items", Element())
    return Array(**{**keyword_filter(Array)(schema), "items": items})


def parse_items(schema: Dict[str, Any], state: ParseState = None) -> Element:
    """Parse array items keyword to DSL Element.

    If not present, defaults to `Element()`.
    """
    state = state or ParseState()
    items = schema.get("items", {})
    if isinstance(items, list):
        raise FeatureNotImplementedError.tuple_array_items()
    return parse_element(items, state)


def _compose_elements(
    element_type: Type[CompositionElement], elements: List[Element]
) -> Element:
    """Create a composition element from a type and list of component elements.

    Filters out trivial elements, and simplifies compositions with only one
    composed element.
    """
    elements = [elem for elem in elements if elem != Element()]
    if not elements:
        return Element()
    if len(elements) == 1:
        return elements[0]
    return element_type(*elements)


def keyword_filter(type_: Type) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    """Create a filter to pull out only relevant keywords for a given type."""
    params = inspect.signature(type_.__init__).parameters.values()
    args = {param.name for param in params}

    def _filter(schema: Dict[str, Any]) -> Dict[str, Any]:
        return {key: value for key, value in schema.items() if key in args}

    return _filter


def _title_format(name: str) -> str:
    """Convert titles in schemas to class names."""
    words = list(filter(None, re.split(r"[ _-]", name)))
    segments = chain.from_iterable(
        [
            re.findall("[A-Z][^A-Z]*", word[0].upper() + word[1:])
            for word in words
        ]
    )
    return "".join(segment.title() for segment in segments)



from typing import Any, Dict, List, Optional, Tuple, Type, Union

from statham.dsl.elements.base import Element
from statham.dsl.helpers import custom_repr
from statham.dsl.property import _Property
from statham.dsl.constants import NotPassed
from statham.dsl.exceptions import SchemaDefinitionError
from statham.dsl.validation import (
    AdditionalProperties,
    InstanceOf,
    Required,
    Validator,
)


class ObjectOptions:

    additionalProperties: Optional[Union[Element, bool]]

    def __init__(self, *, additionalProperties: Union[Element, bool] = True):
        # Name used to match JSONSchema.
        # pylint: disable=invalid-name
        self.additionalProperties = additionalProperties

    def __eq__(self, other) -> bool:
        if not isinstance(other, ObjectOptions):
            return False
        return self.additionalProperties == other.additionalProperties

    def __repr__(self):
        return custom_repr(self)


RESERVED_PROPERTIES = dir(object) + [
    "default",
    "options",
    "properties",
    "additional_properties",
]


class ObjectClassDict(dict):
    """Overriden class dictionary for the metaclass of Object.

    Collects schema properties and default value if present.
    """

    default: Any
    properties: Dict[str, _Property]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.properties = {}
        self.default = self.get("default", NotPassed())
        self.options = self.get("options", ObjectOptions())

    def __setitem__(self, key, value):
        if key in RESERVED_PROPERTIES and isinstance(value, _Property):
            raise SchemaDefinitionError.reserved_attribute(key)
        if key == "default":
            self.default = value
        if key == "options":
            self.options = value
        if isinstance(value, _Property):
            value.bind_name(key)
            return self.properties.__setitem__(key, value)
        return super().__setitem__(key, value)


class ObjectMeta(type, Element):
    """Metaclass to allow declaring Object schemas as classes.

    Collects default value and properties defined as class variables,
    and binds information to those properties.
    """

    properties: Dict[str, _Property]
    options: ObjectOptions

    @staticmethod
    def __subclasses__():
        # This is overriden to prevent errors.
        # TODO: Is there a more elegant way to achieve this? Perhaps
        #   __init_subclass__ should error to prevent this from being
        #   wrong.
        return []

    @classmethod
    def __prepare__(mcs, _name, _bases):
        return ObjectClassDict()

    def __new__(mcs, name: str, bases: Tuple[Type], classdict: ObjectClassDict):
        cls: Type[ObjectMeta] = type.__new__(mcs, name, bases, dict(classdict))
        cls.properties = classdict.properties
        for property_ in cls.properties.values():
            property_.bind_class(cls)
        cls.default = classdict.default
        cls.options = classdict.options
        return cls

    def __hash__(cls):
        return hash(tuple([cls.__name__] + [prop for prop in cls.properties]))

    @property
    def annotation(cls) -> str:
        return cls.__name__

    def __repr__(cls):
        return cls.__name__

    @property
    def type_validator(cls) -> Validator:
        return InstanceOf(dict, cls)

    @property
    def validators(cls) -> List[Validator]:
        return [
            cls.type_validator,
            Required(
                [
                    prop.name
                    for prop in cls.properties.values()
                    if prop.required and not prop.element.default
                ]
            ),
            AdditionalProperties(
                {prop.source for prop in cls.properties.values()},
                cls.options.additionalProperties,
            ),
        ]

    def construct_additional(cls, name: str, value: Any) -> Any:
        """Construct properties not specified in `properties`."""
        element: Element = Element()
        if isinstance(cls.options.additionalProperties, Element):
            element = cls.options.additionalProperties
        additional_property = _Property(element)
        additional_property.bind_class(cls)
        additional_property.bind_name(name)
        return additional_property(value)

    def python(cls) -> str:
        super_cls = next(iter(cls.mro()[1:]))
        class_def = f"""class {repr(cls)}({super_cls.__name__}):
"""
        if (
            not cls.properties
            and isinstance(cls.default, NotPassed)
            and cls.options == ObjectOptions()
        ):
            class_def = (
                class_def
                + """
    pass
"""
            )
        if not isinstance(cls.default, NotPassed):
            class_def = (
                class_def
                + f"""
    default = {repr(cls.default)}
"""
            )
        if cls.options != ObjectOptions():
            class_def = class_def + (
                f"""
    options = {repr(cls.options)}
"""
            )
        for property_ in cls.properties.values():
            class_def = (
                class_def
                + f"""
    {property_.python()}
"""
            )
        return class_def



from typing import List, TypeVar

from statham.dsl.elements.base import Element
from statham.dsl.constants import Maybe, NotPassed
from statham.dsl.validation import InstanceOf


Item = TypeVar("Item")


class Array(Element[List[Item]]):
    """JSONSchema array element.

    Requires schema element for "items" keyword as first positional
    argument. Supported validation keywords provided via keyword arguments.
    # TODO: tuple items
    # TODO: additionalItems
    # TODO: unqiueItems
    # TODO: contains
    """

    items: Element[Item]

    def __init__(
        self,
        items: Element[Item],
        *,
        default: Maybe[List[Item]] = NotPassed(),
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name
        minItems: Maybe[int] = NotPassed(),
        maxItems: Maybe[int] = NotPassed(),
    ):
        self.items = items
        self.default = default
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name
        self.minItems = minItems
        self.maxItems = maxItems

    @property
    def annotation(self) -> str:
        return f"List[{self.items.annotation}]"

    @property
    def type_validator(self):
        return InstanceOf(list)

    def construct(self, value, property_):
        return [
            self.items(item, property_.evolve(property_.name + f"[{idx}]"))
            for idx, item in enumerate(value)
        ]



from typing import Any, ClassVar, Dict

from statham.dsl.elements.meta import ObjectMeta, ObjectOptions
from statham.dsl.property import _Property, UNBOUND_PROPERTY
from statham.dsl.constants import NotPassed


class Object(metaclass=ObjectMeta):
    """Base model for JSONSchema objects.

    New object schemas are defined by implementing subclasses of Object.

    For example:
    ```python
    from statham.dsl.elements import Object, String
    from statham.dsl.property import Property

    class Poll(Object):

        questions: List[str] = Property(String(), required=True)

    poll = Poll({"questions": ["What's up?"]})
    ```
    # TODO: patternProperties
    # TODO: propertyNames
    # TODO: minProperties
    # TODO: maxProperties
    """

    properties: ClassVar[Dict[str, _Property]]
    default: ClassVar[Any]
    options: ClassVar[ObjectOptions]
    additional_properties: Dict[str, Any]

    def __new__(
        cls, value: Any = NotPassed(), property_: _Property = UNBOUND_PROPERTY
    ):
        for validator in cls.validators:
            validator(value, property_)
        if isinstance(value, cls):
            return value
        if isinstance(value, NotPassed) and isinstance(cls.default, NotPassed):
            raise TypeError(f"Missing value argument to {cls}.")
        return object.__new__(cls)

    def __init__(
        self, value: Any = NotPassed(), _property: _Property = UNBOUND_PROPERTY
    ):
        """Initialise the object."""
        if value is self:
            return
        if isinstance(value, NotPassed) and not isinstance(
            self.default, NotPassed
        ):
            value = self.default
        self.additional_properties = {}
        for attr_name, property_ in self.properties.items():
            setattr(
                self,
                attr_name,
                property_(value.pop(property_.source, NotPassed())),
            )
        for name, argument in value.items():
            self.additional_properties[name] = type(self).construct_additional(
                name, argument
            )

    def __repr__(self):
        attr_values = {
            attr: getattr(self, attr) for attr in type(self).properties
        }
        attr_repr = ", ".join(
            [
                f"{attr}={repr(value)}"
                for attr, value in attr_values.items()
                if not isinstance(value, NotPassed)
            ]
        )
        return f"{self.__class__.__name__}({attr_repr})"



from typing import TypeVar, Union

from statham.dsl.elements.base import Element
from statham.dsl.constants import Maybe, NotPassed
from statham.dsl.validation import InstanceOf


T = TypeVar("T", int, float)

Numeric = Union[int, float]


class NumericElement(Element[T]):
    """JSONSchema base numeric element.

    Provides supported validation settings via keyword arguments.
    """

    def __init__(
        self,
        *,
        default: Maybe[T] = NotPassed(),
        minimum: Maybe[Numeric] = NotPassed(),
        maximum: Maybe[Numeric] = NotPassed(),
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name
        exclusiveMinimum: Maybe[Numeric] = NotPassed(),
        exclusiveMaximum: Maybe[Numeric] = NotPassed(),
        multipleOf: Maybe[Numeric] = NotPassed(),
    ):
        self.default = default
        self.minimum = minimum
        self.maximum = maximum
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name
        self.exclusiveMinimum = exclusiveMinimum
        self.exclusiveMaximum = exclusiveMaximum
        self.multipleOf = multipleOf


class Integer(NumericElement[int]):
    """JSONSchema integer element.

    Provides supported validation settings via keyword arguments.
    """

    @property
    def type_validator(self):
        return InstanceOf(int)


class Number(NumericElement[float]):
    """JSONSchema number element.

    Provides supported validation settings via keyword arguments.
    """

    @property
    def type_validator(self):
        return InstanceOf(float)



from itertools import chain
from typing import Dict, Iterator, List, Set, Union

from statham.dsl.constants import Maybe, NotPassed
from statham.dsl.elements import CompositionElement, Element
from statham.dsl.elements.meta import ObjectMeta
from statham.dsl.exceptions import SchemaParseError
from statham.dsl.property import _Property


class ClassDef:
    """Data class for capturing dependencies between Object elements."""

    def __init__(self, element: ObjectMeta, depends: Set[str]):
        self.element: ObjectMeta = element
        self.depends: Set[str] = depends


def _get_dependent_object_elements(element: Maybe[Element]) -> List[ObjectMeta]:
    """Extract any elements which are required by this element.

    Recurses through the schema tree to extract both explicit and
    implicit dependencies.
    """
    if isinstance(element, ObjectMeta):
        return [element]
    if isinstance(element, CompositionElement):
        return list(
            chain.from_iterable(
                _get_dependent_object_elements(sub_element)
                for sub_element in element.elements
            )
        )
    get_keyword = lambda kw: getattr(element, kw, NotPassed())
    items: Maybe[Element] = get_keyword("items")
    properties: Maybe[Dict[str, _Property]] = get_keyword("properties")
    additional_properties: Maybe[Union[Element, bool]] = get_keyword(
        "additionalProperties"
    )
    dependent: Set[ObjectMeta] = set()
    if not isinstance(items, NotPassed):
        dependent = dependent | set(_get_dependent_object_elements(items))
    if not isinstance(properties, NotPassed):
        dependent = dependent | {
            dep
            for prop in properties.values()  # pylint: disable=no-member
            for dep in _get_dependent_object_elements(prop.element)
        }
    if not isinstance(additional_properties, (NotPassed, bool)):
        dependent = dependent | set(
            _get_dependent_object_elements(additional_properties)
        )
    return sorted(dependent, key=lambda obj_type: obj_type.__name__)


def _iter_object_deps(object_type: ObjectMeta) -> Iterator[Element]:
    """Iterate over related elements to an Object subclass."""
    for prop in object_type.properties.values():
        yield prop.element
    if isinstance(object_type.options.additionalProperties, Element):
        yield object_type.options.additionalProperties


class Orderer:
    """Iterator which returns object elements in declaration order.

    Used by the serializer to generate code in the correct order.
    """

    def __init__(self, *elements: Element):
        self._class_defs: Dict[str, ClassDef] = {}
        try:
            self._extract_all(*elements)
        except RecursionError:
            raise SchemaParseError.unresolvable_declaration()

    def _extract_all(self, *elements: Element):
        for element in elements:
            for object_element in _get_dependent_object_elements(element):
                self._extract_elements(object_element)

    def _extract_elements(self, object_type: ObjectMeta) -> Set[str]:
        if object_type.__name__ in self._class_defs:
            return self[object_type.__name__].depends
        deps = {object_type.__name__}
        for sub_element in _iter_object_deps(object_type):
            next_objects = _get_dependent_object_elements(sub_element)
            deps = deps | set(
                chain.from_iterable(map(self._extract_elements, next_objects))
            )
        self._add(
            object_type.__name__,
            ClassDef(
                element=object_type, depends=deps - {object_type.__name__}
            ),
        )
        return deps

    def _add(self, key: str, class_def: ClassDef):
        self._class_defs[key] = class_def

    def _next_key(self) -> str:
        try:
            return next(
                key
                for key, value in self._class_defs.items()
                if not value.depends
            )
        except StopIteration as exc:
            if not self._class_defs:
                raise exc
            raise SchemaParseError.unresolvable_declaration()

    def __getitem__(self, key: str) -> ClassDef:
        return self._class_defs[key]

    def __iter__(self) -> "Orderer":
        return self

    def __next__(self) -> ObjectMeta:
        next_item = self._next_key()
        for value in self._class_defs.values():
            value.depends = value.depends - {next_item}
        return self._class_defs.pop(next_item).element



from statham.dsl.elements.base import Element
from statham.dsl.constants import Maybe, NotPassed
from statham.dsl.validation import InstanceOf


class Boolean(Element[bool]):
    """JSONSchema boolean element."""

    def __init__(self, *, default: Maybe[bool] = NotPassed()):
        self.default = default

    @property
    def type_validator(self):
        return InstanceOf(bool)



from typing import Iterator, Type

from statham.dsl.validation.array import MinItems, MaxItems
from statham.dsl.validation.base import InstanceOf, Validator
from statham.dsl.validation.numeric import (
    Minimum,
    Maximum,
    ExclusiveMinimum,
    ExclusiveMaximum,
    MultipleOf,
)
from statham.dsl.validation.object import AdditionalProperties, Required
from statham.dsl.validation.string import MinLength, MaxLength, Format, Pattern


def _all_subclasses(klass: Type):
    """Get all explicit and implicit subclasses of a type."""
    return set(klass.__subclasses__()).union(
        [s for c in klass.__subclasses__() for s in _all_subclasses(c)]
    )


def get_validators(element) -> Iterator[Validator]:
    """Iterate all applicable validators for an DSL Element.

    Validators identify whether they are applicable for an element via the
    `from_element` class method. In general, this checks whether its
    parameters are present on the element with correct values.
    """
    for validator_type in _all_subclasses(Validator):
        if validator_type is InstanceOf:
            continue
        validator = validator_type.from_element(element)
        if validator:
            yield validator



from typing import Any, Generic, Optional, TypeVar

from statham.dsl.constants import NotPassed
from statham.dsl.elements.base import Element
from statham.dsl.helpers import custom_repr_args


PropType = TypeVar("PropType")


class _Property(Generic[PropType]):
    """Descriptor for a property on an object.

    Used to bind information about the enclosing object to a DSL Element,
    e.g. "required". Should be instantiated via `Property` to enable
    type inference on instances.
    """

    required: bool
    parent: Any
    element: Element
    source: Optional[str]
    name: Optional[str]

    def __init__(
        self,
        element: Element[PropType],
        *,
        required: bool = False,
        source: str = None,
    ):
        self.element = element
        self.required = required
        self.name: Optional[str] = None
        self.source: Optional[str] = source
        self.parent = None

    def __eq__(self, other):
        if not isinstance(other, _Property):
            return False
        return (
            self.element == other.element
            and self.required == other.required
            and self.source == other.source
        )

    def evolve(self, name: str) -> "_Property":
        """Generate renamed property object to pass into nested elements."""
        property_: _Property[PropType] = _Property(
            element=self.element, required=self.required, source=self.source
        )
        property_.bind_name(name)
        property_.bind_class(self.parent)
        return property_

    def bind_name(self, name: str):
        if not self.source:
            self.source = name
        self.name = name

    def bind_class(self, parent: Any):
        self.parent = parent

    def __call__(self, value):
        if not isinstance(self.element.default, NotPassed) and isinstance(
            value, NotPassed
        ):
            value = self.element.default
        if isinstance(value, NotPassed):
            return value
        return self.element(value, self)

    def __repr__(self):
        repr_args = custom_repr_args(self)
        if self.source == self.name:
            _ = repr_args.kwargs.pop("source")
        return f"{self.__class__.__name__}{repr(repr_args)}"

    @property
    def annotation(self):
        if self.required or not isinstance(
            getattr(self.element, "default", NotPassed()), NotPassed
        ):
            return self.element.annotation
        return f"Maybe[{self.element.annotation}]"

    def python(self) -> str:
        prop_def = repr(self).replace(type(self).__name__, Property.__name__)
        return (
            (f"{self.name}: {self.annotation} = {prop_def}")
            if self.name
            else prop_def
        )


UNBOUND_PROPERTY: _Property = _Property(Element(), required=False)
UNBOUND_PROPERTY.bind_name("<unbound>")
UNBOUND_PROPERTY.bind_class(Element())


# Behaves as a wrapper for the `_Property` class.
# pylint: disable=invalid-name
def Property(element: Element, *, required: bool = False, source: str = None):
    """Descriptor for adding a property when declaring an object schema model.

    Return value is typed to inform instance-level interface. See type stubs of
    this module for more detail.

    :param element: The JSONSchema Element object accepted by this property.
    :param required: Whether this property is required. If false, then this
        field may be omitted when data is passed to the outer object's
        constructor.
    :param source: The source name of this property. Only necessary if it must
        differ from that of the attribute, for example when the property name
        conflicts with a reserved keyword. For example, to express a property
        called `default`, one could do the following:
        ```python
        class MyObject(Object):

            # Default for the whole object
            default = {"default": "string"}

            # Property called default
            _default: str = Property(String(), source="default")
        ```
    """
    return _Property(element, required=required, source=source)



# False positive. The cycle exists but is avoided by importing last.
# pylint: disable=cyclic-import
from collections import defaultdict
from typing import (
    Any,
    Callable,
    cast,
    DefaultDict,
    Dict,
    List,
    Generic,
    TypeVar,
    Union,
)

from statham.dsl.constants import NotPassed, Maybe
from statham.dsl.helpers import custom_repr
from statham.dsl.validation import get_validators, InstanceOf, Validator


T = TypeVar("T")

Numeric = Union[int, float]


# This emulates the options available to a general JSONSchema object.
# pylint: disable=too-many-instance-attributes
class Element(Generic[T]):
    """Schema element for composing instantiation logic.

    The generic type is bound by subclasses to indicate their return
    type when called.
    # TODO: enum
    # TODO: const
    # TODO: composition?
    """

    # This is how many options there are!
    # pylint: disable=too-many-locals
    def __init__(
        self,
        *,
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name,redefined-builtin
        default: Maybe[Any] = NotPassed(),
        items: Maybe["Element"] = NotPassed(),
        minItems: Maybe[int] = NotPassed(),
        maxItems: Maybe[int] = NotPassed(),
        minimum: Maybe[Numeric] = NotPassed(),
        maximum: Maybe[Numeric] = NotPassed(),
        exclusiveMinimum: Maybe[Numeric] = NotPassed(),
        exclusiveMaximum: Maybe[Numeric] = NotPassed(),
        multipleOf: Maybe[Numeric] = NotPassed(),
        format: Maybe[str] = NotPassed(),
        pattern: Maybe[str] = NotPassed(),
        minLength: Maybe[int] = NotPassed(),
        maxLength: Maybe[int] = NotPassed(),
        required: Maybe[List[str]] = NotPassed(),
        properties: Maybe[Dict[str, "_Property"]] = NotPassed(),
        additionalProperties: Union["Element", bool] = True,
    ):
        # Bad name to match JSONSchema keywords.
        # pylint: disable=invalid-name
        self.default = default
        self.items = items
        self.minItems = minItems
        self.maxItems = maxItems
        self.minimum = minimum
        self.maximum = maximum
        self.exclusiveMinimum = exclusiveMinimum
        self.exclusiveMaximum = exclusiveMaximum
        self.multipleOf = multipleOf
        self.format = format
        self.pattern = pattern
        self.minLength = minLength
        self.maxLength = maxLength
        self.required = required
        self.properties = properties
        if isinstance(self.properties, dict):
            for name, prop in (self.properties or {}).items():
                prop.bind_class(type(self))
                prop.bind_name(name)
                if prop.required:
                    self.required = list(
                        set(cast(List[str], self.required or []) + [name])
                    )
        self.additionalProperties = additionalProperties

    def __repr__(self):
        """Dynamically construct the repr to match value instantiation."""
        return custom_repr(self)

    def __eq__(self, other):
        """Check if two elements are equivalent.

        This just checks the type and public attribute values. Useful for
        testing parser logic, and could be used to automatically DRY up
        messy schemas in future.
        """
        if not isinstance(other, self.__class__):
            return False
        pub_vars = lambda x: {
            k: v for k, v in vars(x).items() if not k.startswith("_")
        }
        return pub_vars(self) == pub_vars(other)

    @property
    def annotation(self) -> str:
        generic = type(self).__orig_bases__[0].__args__[0]  # type: ignore
        if isinstance(generic, TypeVar):  # type: ignore
            return "Any"
        return generic.__name__

    @property
    def type_validator(self) -> Validator:
        return InstanceOf()

    @property
    def validators(self) -> List[Validator]:
        validators: List[Validator] = [self.type_validator] + list(
            get_validators(self)
        )
        return validators

    def construct(self, value, _property):
        properties = getattr(self, "properties", NotPassed())
        additional_properties = getattr(
            self, "additionalProperties", NotPassed()
        )
        items = getattr(self, "items", NotPassed())
        return object_constructor(properties, additional_properties)(
            array_constructor(items)(value, _property)
        )

    def __call__(self, value, property_=None) -> Maybe[T]:
        """Validate and convert input data against the element.

        Runs validators defined on the `validators` property, and calls
        `construct` on the input.
        """
        property_ = property_ or UNBOUND_PROPERTY
        if not isinstance(self.default, NotPassed) and isinstance(
            value, NotPassed
        ):
            value = self.default
        for validator in self.validators:
            validator(value, property_)
        if isinstance(value, NotPassed):
            return value
        return self.construct(value, property_)


# Needs to be imported last to prevent cyclic import.
# pylint: disable=wrong-import-position
from statham.dsl.property import _Property, UNBOUND_PROPERTY


class _AnonymousObject(dict):
    """Anonymous object type.

    Allows dictionaries passed to untyped objects to use attribute access,
    to match `Object` instances.
    """

    def __getattr__(self, key: str) -> Any:
        return self.__getitem__(key)

    def __setattr__(self, key: str, value: Any):
        return self.__setitem__(key, value)


def object_constructor(
    properties: Maybe[Dict[str, _Property]] = NotPassed(),
    additional_properties: Maybe[Union[Element, bool]] = NotPassed(),
) -> Callable[[Any], _AnonymousObject]:
    """Recursively construct and validate sub-properties of object input."""
    # TODO: Use same `additional_properties` interface as on `Object`.
    default_prop: _Property = _Property(Element())
    if isinstance(additional_properties, Element):
        default_prop = _Property(additional_properties)
    default_properties: DefaultDict[str, _Property] = defaultdict(
        lambda: default_prop
    )
    if not isinstance(properties, NotPassed):
        default_properties.update(properties)

    def _constructor(value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        return _AnonymousObject(
            {
                key: default_properties[key](sub_value)
                for key, sub_value in value.items()
            }
        )

    return _constructor


def array_constructor(
    items: Maybe[Element] = NotPassed()
) -> Callable[[Any, _Property], _AnonymousObject]:
    """Recursively construct and validate sub-properties of array input."""
    items_element: Element = Element()
    if isinstance(items, Element):
        items_element = items

    def _constructor(value: Any, property_: _Property) -> Any:
        if not isinstance(value, list):
            return value
        return [
            items_element(
                item, property_.evolve(property_.name or "" + f"[{idx}]")
            )
            for idx, item in enumerate(value)
        ]

    return _constructor



from typing import Any

from statham.dsl.exceptions import ValidationError
from statham.dsl.validation.base import Validator


class MinItems(Validator):
    types = (list,)
    keywords = ("minItems",)
    message = "Must contain at least {minItems} items."

    def validate(self, value: Any):
        if len(value) < self.params["minItems"]:
            raise ValidationError


class MaxItems(Validator):
    types = (list,)
    keywords = ("maxItems",)
    message = "Must contain fewer than {maxItems} items."

    def validate(self, value: Any):
        if len(value) > self.params["maxItems"]:
            raise ValidationError



from typing import List, Union

from statham.dsl.constants import Maybe
from statham.dsl.elements import (
    AnyOf,
    Array,
    Boolean,
    Integer,
    Null,
    Number,
    OneOf,
    Object,
    ObjectOptions,
    String,
)
from statham.dsl.property import Property


class Category(Object):

    required_name: str = Property(String(), required=True)


class NestedRemote(Object):

    name: Maybe[str] = Property(String())


class Remote(Object):

    name: Maybe[str] = Property(String())

    nested: Maybe[NestedRemote] = Property(NestedRemote)


class Model(Object):

    filesystem_remote_ref_flat: Maybe[Category] = Property(Category)

    filesystem_remote_ref_directory: Maybe[Remote] = Property(Remote)



from typing import List, Union

from statham.dsl.constants import Maybe
from statham.dsl.elements import (
    AnyOf,
    Array,
    Boolean,
    Integer,
    Null,
    Number,
    OneOf,
    Object,
    ObjectOptions,
    String,
)
from statham.dsl.property import Property


class Category(Object):

    required_name: str = Property(String(), required=True)


class Child(Object):

    name: Maybe[str] = Property(String())

    category: Maybe[Category] = Property(Category)


class Model(Object):

    children: Maybe[List[Child]] = Property(Array(Child))

    category: Maybe[Category] = Property(Category)



from typing import Dict

from statham.dsl.constants import JSONElement


class StathamError(Exception):
    """Base exception"""


class SchemaDefinitionError(StathamError):
    """Failure when declaring schemas using DSL."""

    @classmethod
    def reserved_attribute(cls, attribute_name: str) -> "SchemaDefinitionError":
        # TODO: Default to underscore after not before.
        return cls(
            f"May not use reserved attribute `{attribute_name}` as a property "
            "attribute name. Instead use "
            f"`_{attribute_name} = Property(<element>, "
            f"source='{attribute_name}'`"
        )


class ValidationError(StathamError):
    """Validation failure in generated models."""

    @classmethod
    def from_validator(cls, property_, value, message) -> "ValidationError":
        return cls(
            f"Failed validating `{repr(property_.parent)}."
            f"{property_.name} = {repr(value)}`. {message}"
        )

    @classmethod
    def combine(
        cls, property_, value, exceptions, message
    ) -> "ValidationError":
        base_message = str(cls.from_validator(property_, value, ""))
        error_breakdown = ", ".join(str(exc) for exc in exceptions)
        error_breakdown = error_breakdown.replace(base_message, "")
        return cls.from_validator(
            property_, value, message + f" Individual errors: {error_breakdown}"
        )

    @classmethod
    def multiple_composition_match(cls, matching_models, data):
        return cls(
            "Matches multiple possible models. Must only match one.\n"
            f"Data: {data}\n"
            f"Models: {matching_models}"
        )


class SchemaParseError(StathamError):
    """Failure during parsing of provided JSON Schema input."""

    @classmethod
    def missing_title(
        cls, schema: Dict[str, JSONElement]
    ) -> "SchemaParseError":
        return cls(
            "No title defined in schema. Use "
            "`statham.titles.title_labeller` to pre-process the "
            f"schema: {schema}"
        )

    @classmethod
    def unresolvable_declaration(cls) -> "SchemaParseError":
        return cls(
            "Schema document has an unresolvable declaration tree. This "
            "generally occurs due to cyclical references."
        )

    @classmethod
    def invalid_type(cls, value):
        return cls(f"Got invalid type keyword: {value}.")


class FeatureNotImplementedError(SchemaParseError):
    """Functionality not yet implemented."""

    @classmethod
    def tuple_array_items(cls) -> "FeatureNotImplementedError":
        return cls("Tuple array items are not supported.")



from statham.dsl.elements.base import Element
from statham.dsl.constants import Maybe, NotPassed
from statham.dsl.validation import InstanceOf


class Null(Element[None]):
    """JSONSchema null element."""

    def __init__(self, *, default: Maybe[None] = NotPassed()):
        self.default = default

    @property
    def annotation(self) -> str:
        return "None"

    @property
    def type_validator(self):
        return InstanceOf(type(None))
